from typing import Optional

from webpilot.config import WebPilotConfig
from webpilot import models
from pydantic import parse_obj_as
from websocket import create_connection, WebSocket
from pathlib import Path
import sys
import subprocess
import urllib.request
import json
import time
from bs4 import BeautifulSoup

is_windows = sys.platform.startswith('win')


def _send_message(id: int, ws: WebSocket, expected: models.Model, method: str, session_id: Optional[str], payload: dict) -> models.Model:
    msg = {'id': id, 'method': method, 'params': payload}
    if session_id is not None:
        msg['sessionId'] = session_id
    ws.send(json.dumps(msg))
    while True:
        return_value: dict = json.loads(ws.recv())
        if return_value.get('id', None) == None:
            continue

        error_value = return_value.get('error', None)
        if error_value != None:
            raise error_value

        result_value: Optional[dict] = return_value.get('result', None)
        if result_value == None:
            continue

        return parse_obj_as(expected, result_value)


class Robot:
    def __init__(self, process: subprocess.Popen, endpoint: str, port: int):
        self._process = process
        self._counter = 0
        self._endpoint = endpoint
        self._port = port

    def _send_message(self, expected: models.Model, method: str, session_id: Optional[str], payload: dict) -> models.Model:
        self._counter = self._counter + 1
        return _send_message(self._counter, self._ws, expected,
                             method, session_id, payload)

    def new_tab(self):
        """Open a tab in the browser"""
        target = self._send_message(models.Target, "Target.createTarget", None, {
            'url': 'about:blank'
        })
        result = self._send_message(models.Target, "Target.attachToTarget", None, {
            'targetId': target.targetId,
            'flatten': True
        })
        result.targetId = target.targetId
        return result

    def navigate(self, target: models.Target, url: str, delay: int = 1):
        """navigate to a url an await delay seconds before return the page"""
        page = self._send_message(models.Page, "Page.navigate", target.sessionId, {
            'url': url
        })
        page.target = target
        time.sleep(delay)
        return page

    def get_content(self, page: models.Page):
        document = self._send_message(
            models.Document, 'DOM.getDocument', page.target.sessionId, {})
        content = self._send_message(models.OuterHTML, 'DOM.getOuterHTML', page.target.sessionId, {
            'nodeId': document.root.nodeId
        })
        return content.outerHTML

    def get_dom(self, page: models.Page):
        content = self.get_content(page)
        return BeautifulSoup(content)

    def close(self):
        self._ws.close()
        self._process.kill()

    def __enter__(self):
        self._ws = create_connection(self._endpoint)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def open_chrome(config: WebPilotConfig = WebPilotConfig()):
    arguments = [config.chrome_executable]
    arguments.append(f'--remote-debugging-port={config.remote_port}')

    if config.headless:
        arguments.append('--headless')
        if is_windows:
            arguments.append('--disable-gpu')
    else:
        Path('.webrobot_data').mkdir(exist_ok=True)
        arguments.append(
            "--no-default-browser-check --user-data-dir=.webrobot_data")

    arguments.append(config.url)

    process = subprocess.Popen(arguments)

    time.sleep(1)

    with urllib.request.urlopen(f'http://localhost:{config.remote_port}/json/version') as response:
        metadata: dict = json.loads(response.read())
        endpoint = metadata.get('webSocketDebuggerUrl')

    return Robot(process=process, endpoint=endpoint, port=config.remote_port)



from typing import Optional, TypeVar
from pydantic import BaseModel

Model = TypeVar('Model')


class Target(BaseModel):
    targetId: Optional[str]
    sessionId: Optional[str]


class Page(BaseModel):
    target: Optional[Target]
    frameId: str
    errorText: Optional[str]


class Node(BaseModel):
    nodeId: int


class Document(BaseModel):
    root: Node


class OuterHTML(BaseModel):
    outerHTML: str



from pydantic import BaseModel
import socket


def _get_free_tcp_port() -> int:
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(('', 0))
    _, port = tcp.getsockname()
    tcp.close()
    return port


class WebPilotConfig(BaseModel):
    remote_port: int = _get_free_tcp_port()
    url: str = "about:blank"
    headless: bool = True
    chrome_executable: str = 'chrome'



from setuptools import find_packages, setup
import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / 'README.md').read_text()

setup(
    name='webpilot',
    packages=find_packages(include=['webpilot']),
    version='0.1.0',
    description='Api to control chromium browsers using puppeteer',
    long_description=README,
    long_description_content_type='text/markdown',
    author='fernandojerez',
    license='Apache',
    requires=['websocket_client', 'pydantic', 'beautifulsoup4'],
    install_requires=["websocket_client", 'pydantic', 'beautifulsoup4'],
    test_suite='tests'
)
